/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * Translates the DAG (parse tree) to circuit.
 *
 * Marcio Oliveira -- 2002
 */

#include <translation.h>
#include <error.h>
#include <hash.h>

controlPointPtr controlPointList = NULL;
int controlPointCount = 0;

actionPointPtr actionPointList = NULL;
int actionPointCount = 0;  

exThreadPtr exThreadList = NULL;
int exThreadCount = 0;

wirePtr wireList = NULL;
int wireCount = 0;

treeNodePtr build(treeNodePtr node, treeNodePtr f, treeNodePtr ffc,
                  treeNodePtr *ok, treeNodePtr *okp, treeNodePtr *en,
                  treeNodePtr *okfc, treeNodePtr *acfc, treeNodePtr *enfc) {

    controlPointPtr cp;
    treeNodePtr temp[7];
    treeNodePtr tOk[4];
    treeNodePtr tOkp[4];
    treeNodePtr tEn[4];
    treeNodePtr tOkfc[4];
    treeNodePtr tAcfc[4];
    treeNodePtr tEnfc[4];
    treeNodePtr h;
    static int index = 0;
    static treeNodePtr false = NULL;

    if (false == NULL)
            false = createTreeNode(NULL, NULL, P_TN_FALSE_C, NULL, NULL);

    if (isBoolExpr(node)) {
        cp = createControlPoint();
        cp->f = createTreeNode(NULL, NULL, P_TN_AND_C, node, f);
        temp[0] = createTreeNode(cp, NULL, P_TN_CP, NULL, NULL);
        if (node->actionList != NULL) {
            createActionPoint(node->actionList, temp[0]);
        } 

        *en = createWireNode(f);
        *ok = createWireNode(cp->f);
        *okp = *ok;
     
        if (ffc == false)
            *okfc = false;
        else  
            *okfc = createTreeNode(NULL, NULL, P_TN_AND_C, node, ffc);

        *acfc = false;
        *enfc = ffc;
         
        return temp[0];
    }

    switch (node->type) {
        case P_TN_OR_P:
            temp[0] = build(node->left, f, ffc, &tOk[0], &tOkp[0], &tEn[0], &tOkfc[0], &tAcfc[0], &tEnfc[0]);

            temp[1] = build(node->right, f, ffc, &tOk[1], &tOkp[1], &tEn[1], &tOkfc[1], &tAcfc[1], &tEnfc[1]);

            h = createTreeNode(NULL, NULL, P_TN_OR_C, temp[0], temp[1]);

            *en = createTreeNode(NULL, NULL, P_TN_OR_C, tEn[0], tEn[1]);
            *ok = createTreeNode(NULL, NULL, P_TN_OR_C, tOk[0], tOk[1]);
            *okp = createTreeNode(NULL, NULL, P_TN_OR_C, tOkp[0], tOkp[1]);

            *okfc = createTreeNode(NULL, NULL, P_TN_OR_C, tOkfc[0], tOkfc[1]);
            *acfc = createTreeNode(NULL, NULL, P_TN_OR_C, tAcfc[0], tAcfc[1]);
            *enfc = createTreeNode(NULL, NULL, P_TN_OR_C, tEnfc[0], tEnfc[1]);

            *en = createWireNode(*en);
            *ok = createWireNode(*ok);
            *okp = createWireNode(*okp);

            *acfc = createWireNode(*acfc);
            *okfc = createWireNode(*okfc);
            *enfc = createWireNode(*enfc); 
 
            break;
        case P_TN_COMMA_P:
            temp[0] = build(node->left, f, ffc, &tOk[0], &tOkp[0], &tEn[0], &tOkfc[0], &tAcfc[0], &tEnfc[0]);

            h = build(node->right, temp[0], tAcfc[0], &tOk[1], &tOkp[1], &tEn[1], &tOkfc[1], &tAcfc[1], &tEnfc[1]);

            *en = createTreeNode(NULL, NULL, P_TN_OR_C, tEn[0], tEn[1]);

            temp[1] = createTreeNode(NULL, NULL, P_TN_NOT_C, tEn[0], NULL);
            temp[1] = createTreeNode(NULL, NULL, P_TN_OR_C, temp[1], tOk[0]); 
 
            temp[2] = createTreeNode(NULL, NULL, P_TN_NOT_C, tEn[1], NULL);
            temp[2] = createTreeNode(NULL, NULL, P_TN_OR_C, temp[2], tOk[1]);
             
            temp[3] = createTreeNode(NULL, NULL, P_TN_OR_C, tOk[0], tOk[1]);

            *ok = createTreeNode(NULL, NULL, P_TN_AND_C, temp[1], temp[2]);
            *ok = createTreeNode(NULL, NULL, P_TN_AND_C, *ok, temp[3]);

            *okp = createTreeNode(NULL, NULL, P_TN_OR_C, tOkp[0], tOkp[1]);

            temp[4] = createTreeNode(NULL, NULL, P_TN_NOT_C, tEnfc[0], NULL);
            temp[4] = createTreeNode(NULL, NULL, P_TN_OR_C, temp[4], tOkfc[0]);

            temp[5] = createTreeNode(NULL, NULL, P_TN_NOT_C, tEnfc[1], NULL);
            temp[5] = createTreeNode(NULL, NULL, P_TN_OR_C, temp[5], tOkfc[1]);

            temp[6] = createTreeNode(NULL, NULL, P_TN_OR_C, tOkfc[0], tOkfc[1]);

            *okfc = createTreeNode(NULL, NULL, P_TN_AND_C, temp[4], temp[5]);
            *okfc = createTreeNode(NULL, NULL, P_TN_AND_C, *okfc, temp[6]);

            *acfc = createTreeNode(NULL, NULL, P_TN_OR_C, tAcfc[0], tAcfc[1]);

            *enfc = createTreeNode(NULL, NULL, P_TN_OR_C, tEnfc[0], tEnfc[1]);

            *en = createWireNode(*en);
            *ok = createWireNode(*ok);
            *okp = createWireNode(*okp);

            *acfc = createWireNode(*acfc);
            *okfc = createWireNode(*okfc);
            *enfc = createWireNode(*enfc);

            break;
        case P_TN_STAR_P:
            temp[0] = createTreeNode((void *) index++, NULL, P_TN_TEMP, 
                                                                    NULL, NULL);

            temp[3] = createTreeNode((void *) index++, NULL, P_TN_TEMP,
                                                                    NULL, NULL);

            temp[1] = build(node->left, temp[0], temp[3], &tOk[0], &tOkp[0], &tEn[0], &tOkfc[0], &tAcfc[0], &tEnfc[0]);

            h = createTreeNode(NULL, NULL, P_TN_OR_C, temp[1], f);
            temp[0]->left = h;

            *acfc = createTreeNode(NULL, NULL, P_TN_OR_C, tAcfc[0], ffc);
            temp[3]->left = *acfc;

            *enfc = tEnfc[0];

            *en = tEn[0]; 
            *ok = createTreeNode(NULL, NULL, P_TN_OR_C, h, tOk[0]);
            *okp = tOkp[0];

            *okfc = createTreeNode(NULL, NULL, P_TN_OR_C, *acfc, tOkfc[0]);

            temp[2] = createTreeNode(NULL, NULL, P_TN_NOT_C, tOkp[0], NULL);
            h = createTreeNode(NULL, NULL, P_TN_AND_C, temp[2], h);    

            *ok = createWireNode(*ok);
            h = createWireNode(h);

            *acfc = createTreeNode(NULL, NULL, P_TN_AND_C, temp[2], *acfc);

            *okfc = createWireNode(*okfc);
            *acfc = createWireNode(*acfc);

            break;
        case P_TN_AT_P:

            temp[0] = createTreeNode((void *) index++, NULL, P_TN_TEMP,
                                                                    NULL, NULL);

            h = build(node->left, f, ffc, &tOk[0], &tOkp[0], &tEn[0], &tOkfc[0], &tAcfc[0], &tEnfc[0]);
            temp[2] = build(node->right, temp[0], h, &tOk[1], &tOkp[1], &tEn[1], &tOkfc[1], &tAcfc[1], &tEnfc[1]);
            temp[0]->left = h;

            tOk[2] = copyAndSubstitute(tOk[1], temp[0], false);
            tOkp[2] = copyAndSubstitute(tOkp[1], temp[0], false);
            tEn[2] = copyAndSubstitute(tEn[1], temp[0], false);

            tOk[3] = tOkfc[1];
            tEn[3] = h;

            temp[3] = createTreeNode(NULL, NULL, P_TN_NOT_C, tEn[2], NULL);
            temp[3] = createTreeNode(NULL, NULL, P_TN_OR_C, temp[3], tOk[2]);

            temp[4] = createTreeNode(NULL, NULL, P_TN_NOT_C, tEn[3], NULL);
            temp[4] = createTreeNode(NULL, NULL, P_TN_OR_C, temp[4], tOk[3]);

            temp[5] = createTreeNode(NULL, NULL, P_TN_AND_C, temp[4], temp[3]);
            temp[5] = createWireNode(temp[5]);

            createExThread(temp[5], h, tEn[1], tOkp[2]);

            *en = tEn[0];
            *ok = tOk[0];
            *okp = tOkp[0];
            
            *acfc = tAcfc[0];
            *okfc = tOkfc[0];
            *enfc = tEnfc[0]; 

            break;
        default:
            fatal_error("build()");
    }

    if (node->actionList != NULL) {
        createActionPoint(node->actionList, h);
    }

    return h;
}

controlPointPtr createControlPoint() {

    controlPointPtr c;

    if ((c = (controlPointPtr) malloc(sizeof(struct controlPoint))) == NULL) 
        fatal_error("createControlPoint()");

    memset(c, 0, sizeof(struct controlPoint));
    c->index = controlPointCount++;
    c->next = controlPointList;
    controlPointList = c;

    return c;
}

void createActionPoint(actionPtr al, treeNodePtr f) {

    actionPointPtr ap;
    actionPointPtr temp;

    if ((ap = (actionPointPtr) malloc(sizeof(struct actionPoint))) == NULL)
        fatal_error("createActionPoint()");

    memset(ap, 0, sizeof(struct actionPoint));
    ap->actionList = al;
    ap->f = f; 
    ap->index = actionPointCount++; 

    if (actionPointList == NULL) {
        actionPointList = ap;  
    } else {
        temp = actionPointList;
        while (temp->next != NULL) {
            temp = temp->next;
        }
        temp->next = ap;   
    }
}

void createExThread(treeNodePtr o, treeNodePtr a, 
                    treeNodePtr en, treeNodePtr r) {

    exThreadPtr et;

    if ((et = (exThreadPtr) malloc(sizeof(struct exThread))) == NULL) 
        fatal_error("createExThread()"); 

    memset(et, 0, sizeof(struct exThread));
    et->index = exThreadCount++;
    et->ok = o;
    et->active = a;
    et->enable = en;
    et->rec = r;

    et->next = exThreadList;
    exThreadList = et; 
}

treeNodePtr copyAndSubstitute(treeNodePtr or, treeNodePtr old, 
                              treeNodePtr new) {

    treeNodePtr result;
    wirePtr w;

    if (isBoolExpr(or))
        return or;

    switch (or->type) {
        case P_TN_CP:
            result = or;
            break;
        case P_TN_AND_C:
            result = createTreeNode(NULL, NULL, P_TN_AND_C,
                                    copyAndSubstitute(or->left, old, new),
                                    copyAndSubstitute(or->right, old, new));
            break;
        case P_TN_OR_C:
            result = createTreeNode(NULL, NULL, P_TN_OR_C,
                                    copyAndSubstitute(or->left, old, new),
                                    copyAndSubstitute(or->right, old, new));
            break;
        case P_TN_NOT_C:
            result = createTreeNode(NULL, NULL, P_TN_NOT_C,
                                    copyAndSubstitute(or->left, old, new),
                                    NULL);
            break;
        case P_TN_TEMP:
            if (or == old) {
                result = new;
            } else {
                result = copyAndSubstitute(or->left, old, new);
            }
            break;
        case P_TN_TRUE_C:
            result = or;
            break;
        case P_TN_FALSE_C:
            result = or; 
            break;
        case P_TN_WIRE_C:
            w = (wirePtr) or->data;
            result = createWireNode(copyAndSubstitute(w->f, old, new));
            break;
        default:
            fatal_error("copyAndSubstitute()");
    }

    return result;
}


treeNodePtr copyAndDelete(treeNodePtr or, treeNodePtr old,
                              treeNodePtr false) {

    treeNodePtr result;
    wirePtr w;

    if (isBoolExpr(or))
        return or;

    switch (or->type) {
        case P_TN_CP:
            result = false;
            break;
        case P_TN_AND_C:
            result = createTreeNode(NULL, NULL, P_TN_AND_C,
                                    copyAndDelete(or->left, old, false),
                                    copyAndDelete(or->right, old, false));
            break;
        case P_TN_OR_C:
            result = createTreeNode(NULL, NULL, P_TN_OR_C,
                                    copyAndDelete(or->left, old, false),
                                    copyAndDelete(or->right, old, false));
            break;
        case P_TN_NOT_C:
            result = createTreeNode(NULL, NULL, P_TN_NOT_C,
                                    copyAndSubstitute(or->left, old, false),
                                    NULL);
            break;
        case P_TN_TEMP:
            if (or == old) {
                result = or;
            } else {
                result = copyAndDelete(or->left, old, false);
            }
            break;
        case P_TN_TRUE_C:
            result = or;
            break;
        case P_TN_FALSE_C:
            result = or;
            break;
        case P_TN_WIRE_C:
            w = (wirePtr) or->data;
            result = createWireNode(copyAndSubstitute(w->f, old, false));
            break;
        default:
            fatal_error("copyAndSubstitute()");
    }

    return result;
}

int lambda(treeNodePtr node) {

    int r1, r2;
    int result;

    if (isBoolExpr(node)) 
        return 0;

    switch (node->type) {
        case P_TN_OR_P:
            r1 = lambda(node->left);
            r2 = lambda(node->right);
            result = (r1 || r2);
            break;
        case P_TN_COMMA_P:
            r1 = lambda(node->left);
            r2 = lambda(node->right);
            result = (r1 && r2);
            break;
        case P_TN_STAR_P:
            result = 1;
            break;
        case P_TN_AT_P:
            result = lambda(node->left); 
            break;
        default:
            fatal_error("lambda()");
    }

    return result;
}

treeNodePtr createWireNode(treeNodePtr f) {

    wirePtr w;
    static wirePtr wLast = NULL;

    w = (wirePtr) malloc(sizeof(struct wire));
    if (w == NULL)
        fatal_error("createWire()");

    memset(w, 0, sizeof(struct wire));
    
    w->f = f;
    w->index = wireCount++;
    
    if (wireList == NULL) {
        wireList = w;
        wLast = wireList;
    } else {
        wLast->next = w;
        wLast = wLast->next;
    }

    return createTreeNode(w, NULL, P_TN_WIRE_C, NULL, NULL);   
}
